; 1.scm - First set of example Scheme code

( let ((a 12)
       (b 5)
       (square (lambda (x) (* x x)))
       (plus +)
      )
  (sqrt (plus (square a)
              (square b)
        )
  )
)

; let's define a function...
(define double 
  (lambda (x) (* 2 x))
)

(double 55)

; how do you write a recursive function?
; first need to know how to compare (to test for base case)

; This is wrong because there is no base case!
;(define fact
;  (lambda (x) (* x (fact (- x 1))))
;)

; Let's try out "cond"
(cond ((< 5 0) -1)
      ((= 5 0)  0)
      (#t 1)                  ; "#t" handles all other cases
)

(define sign
  (lambda (x) (cond ((< x 0) -1)
                    ((= x 0) 0)
                    (#t 1)
              )
  )
)

(sign 400)
(sign 0)
(sign -333)

; Now we can try factorial
; don't forget to multiply by x
(define fact
  (lambda (x) (cond ((< x 2) 1)
                    (#t (* x (fact (- x 1))))
              )
  )
)

(fact 7)

; function defined for a list.
(define third
  (lambda (x) (caddr x))
)

(third '(alpha beta gamma delta epsilon zeta))

; Note that we need a quote when we want to write a list-constant
;   or atom from a list.  That is, if you say x, this is the 
;   variable x, but if you say 'x this means don't evaluate x.

; Can we define a function on two variables?
(define sum
  (lambda (x y) (+ x y))
)

(sum 8 4)

; Let's add elements in a list of numbers.
; This version is wrong because there's no base case!
(define sumlist
  (lambda (x) (+ (car x) (sumlist (cdr x))))
)

; Correct version
; Note the use of the "null?" built-in function as well as "#t"
(define sumlist
  (lambda (x) (cond ((null? x) 0)
                    (#t (+ (car x) (sumlist (cdr x))))
              )
  )
)

(sumlist '(5 4 7 3 2 6 1 8 9 10))


;   How do we define/create a list?

(append '(a b) '(c d e))
(append '((a b) c) '(d (e f)))

(append '(5) '(6 7))
(append '((a b)) '((i j) (e f)))

(define list1 '(1 2 3))
(append '(8) list1)              ; 8 has to be in a (quoted) list.

(define list2 '((a b) (i j) (e f)))      ; didn't have any output by itself
list2

(car list2)                      ; should be (a b)
(caar list2)                     ; should just be "a"
(cdr list2)

; The following two give error!  How to compare to atom like "i" ?
; ANSWER: use "eq?" instead of "=" when comparing non-numbers
;(= 'i (caar list2))
;(= 'i 'i)

(eq? 'i 'i)
(eq? 'i (caar list2))
(eq? 'i (cadar list2))

; Well, what is the cadar of list2?
(cadar list2)
