; 2.scm - 2nd example source code of Scheme/Lisp. Contents:
; (length L) - # of items in a list
; (number_matches word L) - # times a given word appears in a list
; (get_word L n) - return the n-th word in the list
; (get_last_word L) - calls previous function :)
; (replace_first L word)
; (replace_cp L)
; (replace L victim replacement)
; (wrong_reverse L) - only puts the first word at the end!
; (reverse L)
; (abs L) - find the absolute value of each number in the list
; (get_age R) - defined for property list
; (set_age R new_age) - defined for property list
; (get_salary R) - defined for assocation list
; (set_salary R new_salary) - defined for assocation list
;   Also, doesn't matter if the list parameter is called L or R or ...

(display "hello\n")            ; use this function to display strings

(define length
  (lambda (L)
    (cond ((null? L) 0)
          (#t (+ 1 (length (cdr L))))
    )
  )
)

(define str1 '(There's no limit to what he could do. He could destroy the Earth.))
(length str1)

; count how many times a word appears in a list
; Not case sensitive
(define number_matches
  (lambda (word L)
    (cond ((null? L) 0)
          ((eq? word (car L)) (+ 1 (number_matches word (cdr L))))
          (#t (number_matches word (cdr L)))
    )
  )
)

(number_matches 'he str1)
(number_matches 'nikto str1)

; For discussion:  we have car, so how do we obtain last element?
; More general:  given #, find that element :  4 --> give 4th element.

; Assume the numbering starts at 1
(define get_word
  (lambda (L n)
    (cond ((null? L) 'out-of-bounds)
          ((= n 1) (car L))
          (#t (get_word (cdr L) (- n 1)))
    )
  )
)

; We get to call a function we just defined!
(define get_last_word
  (lambda (L)
    (get_word L (length L))
  )
)

(get_word str1 4)
(get_word str1 13)
(get_word str1 14)
(get_word str1 11)
(get_last_word str1)
          
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Ok let's try something else - change first word of a list
(define replace_first
  (lambda (L word)
    (cond ((null? L) L)
          (#t (cons word (cdr L)))
    )
  )
)

(replace_first '(one of these things) 'five)

; replace coke with pepsi
; originally forgot to call replace_cp in 2nd line of cond
;   which means we only replace first coke
(define replace_cp
  (lambda (L)
    (cond ((null? L) L)
          ((eq? (car L) 'coke) (cons 'pepsi (replace_cp (cdr L))))
          (#t (cons (car L) (replace_cp (cdr L))))
    )
  )
)

(define str2 '(would you like a coke can or just some coke))
(replace_cp str2)

; A generalized replace
; Don't forget when calling ourselves to include proper parameters.
(define replace_word
  (lambda (L victim replacement)
    (cond ((null? L) L)
          ((eq? (car L) victim) 
                 (cons replacement (replace_word (cdr L) victim replacement)))
          (#t 
                 (cons (car L) (replace_word (cdr L) victim replacement)))
    )
  )
)

str2
(replace_word str2 'can 'bottle)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; How do we reverse a list?
;   idea:  use append to put car after the cdr
; This first version is wrong because I forgot recursion!
(define wrong_reverse
  (lambda (L)
    (cond ((null? L) L)
          (#t (append (cdr L) (cons (car L) ())))
    )
  )
)

; Put recursive call around where we have "(cdr L)"
(define reverse
  (lambda (L)
    (cond ((null? L) L)
          (#t (append (reverse (cdr L)) (cons (car L) ())))
    )
  )
)

(define str3 '(it has been observed))
(wrong_reverse str3)
(reverse str3)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Number stuff, like absolute value of each element.
; don't forget recursive call
(define abs
  (lambda (L)
    (cond ((null? L) L)
          ((> (car L) 0) (cons (car L) (abs (cdr L))))
          (#t (cons (- 0 (car L)) (abs (cdr L))))
    )
  )
)

(define list5 '(5 -2 0 -8 4))
(abs list5)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Property lists
(define record1 '(name (John Doe) salary 25000 age 51 hire (June 10 1989)))
(define record2 '(salary 16000 name (Sam Coetzer) hire (May 22 1996) age 30))

; base case - if no age present, return 0
; I had forgotten () around "cddr R" and it complained about too many args
(define get_age
  (lambda (R)
    (cond ((null? R) 0)
          ((eq? (car R) 'age) (cadr R))
          (#t (get_age (cddr R)))             
    )
  )
)

(get_age record1)
(get_age record2)

; base case - no age present, do nothing (return list unchanged)
; If we find "age" at the front of the list, keep "age", and replace
;   the 2nd element with the parameter new_age
; If the front is not "age", then pass by the first 2 elements
;   which are (car R) and (cadr R).  Continue from 3rd with (cddr R).
; Hey, shouldn't I be calling recursively on 3rd line of cond?
;   look here -> (#t (cons (car R) (cons (cadr R)   (cddr R)  )))
;   looks like I forgot "set_age" and the "new_age" parameter.
(define set_age
  (lambda (R new_age)
    (cond ((null? R) R)
          ((eq? (car R) 'age) (cons 'age (cons new_age (cddr R))))
          (#t (cons (car R) (cons (cadr R) (set_age (cddr R) new_age))))
    )
  )
)

(set_age record1 53)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Analogous to property list is assocation list.  Easier to read.
; (database):  we can ask for the value corresponding to a key.
;   or even return all keys that have a corresponding value
; ...
;
; Wouldn't it be a good idea to have a _general_ get and set
;   rather than a specific one for just salary?  --> need 1 more param.

(define record3 '((age 41)(name (Bob Gette))(hire (May 1 1968))(salary 38000)))
(define record4 '((salary 90000)(hire (January 5 2000))(name (Jill Beach))(age 50)))

; If we want to write a function to change age, needs to have same
; name because we already have such a function for property lists.
; So why not just write a function to change _salary_.
; Return 0 if can't find salary key.
(define get_salary
  (lambda (L)
    (cond ((null? L) 0)
          ((eq? 'salary (caar L)) (cadar L))
          (#t (get_salary (cdr L)))
    )
  )
)

(get_salary record3)
(get_salary record4)

; Change salary.  If no salary field, do nothing.
;   (Note - Alternatively, we could create a salary field.)
; Hmmm, setting the 2nd element of inner list (i.e. cadar) not that easy.
;   Empty list is '()
;   Then we need to use cons to put new_salary in:
;                         (cons new_salary ())
;   Then we need to put word "salary" at front:
;           (cons 'salary (cons new_salary ()))
;   And this needs to be the new first element to go before rest of list
;     in place of the current field for salary.
;     (cons (cons 'salary (cons new_salary ())) (cdr L))
;   Fortunately there's only one salary field, so "cdr L" is sufficient
;   The recursive call is when we didn't find salary at front.
;     In this case we do the usual thing of passing by the car L...
;     We need cons to put first (nonsalary) field back into list
;       when the recursion is finished.
(define set_salary
  (lambda (L new_salary)
    (cond ((null? L) L)
          ((eq? 'salary (caar L)) (cons (cons 'salary (cons new_salary ())) (cdr L)))
          (#t (cons (car L) (set_salary (cdr L) new_salary)))
    )
  )
)

(set_salary record3 38500)
(set_salary record4 90500)

; Note that we are not permanently changing salary.
; You get unexpected end-of-file if you are missing ')'
(get_salary (set_salary record4 90500))
(get_salary record4)
